"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const sortBy_1 = tslib_1.__importDefault(require("lodash/sortBy"));
const inquirer_1 = tslib_1.__importDefault(require("inquirer"));
const listr_1 = tslib_1.__importDefault(require("listr"));
const promise_retry_1 = tslib_1.__importDefault(require("promise-retry"));
const errors_1 = tslib_1.__importDefault(require("@tradle/errors"));
const errors_2 = require("./errors");
const logger_1 = require("./logger");
const utils_1 = require("./utils");
const USE_CURRENT_USER_ROLE = true;
const MIN_VERSION = '01.01.0f';
exports.update = (conf, { stackId, tag, provider, showReleaseCandidates, force, }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const getUpdateWithRetry = tag => {
        let requested;
        return promise_retry_1.default((retry, number) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                return yield conf.getUpdateInfo({ tag });
            }
            catch (err) {
                errors_1.default.rethrow(err, ['developer', errors_2.Errors.InvalidInput]);
                if (!requested && errors_1.default.matches(err, errors_2.Errors.NotFound)) {
                    requested = true;
                    yield conf.requestUpdate({ tag, provider });
                }
                retry(err);
            }
        }), {
            maxTimeout: 10000,
            minTimeout: 5000,
            retries: 10
        });
    };
    const version = yield conf.getCurrentVersion();
    if (!(version.sortableTag && version.sortableTag > MIN_VERSION)) {
        logger_1.logger.info(`you have an old version of MyCloud which doesn't support the new update mechanism
Please update manually this one time. See instructions on https://github.com/tradle/serverless`);
        return;
    }
    if (!tag) {
        let updates = yield conf.listUpdates({ provider });
        if (!showReleaseCandidates) {
            updates = updates.filter(update => !isReleaseCandidateTag(update.tag));
        }
        if (!updates.length) {
            logger_1.logger.info(`no updates available`);
            return;
        }
        updates = sortBy_1.default(updates, 'sortableTag');
        const result = yield inquirer_1.default.prompt([{
                type: 'list',
                name: 'tag',
                message: 'Choose a version to update to',
                choices: updates.map(({ tag }) => ({
                    name: getChoiceTextForTag(tag),
                    value: tag
                })),
            }]);
        tag = result.tag;
        if (!force) {
            const idx = updates.findIndex(update => update.tag === tag);
            if (idx !== -1) {
                const transition = updates.slice(0, idx).find(update => isTransitionReleaseTag(update.tag));
                if (transition) {
                    logger_1.logger.info(`you must apply the transition version first: ${transition.tag}`);
                    yield utils_1.confirmOrAbort(`apply transition tag ${transition.tag} now?`);
                    tag = transition.tag;
                }
            }
        }
    }
    const triggerUpdate = (update) => tslib_1.__awaiter(this, void 0, void 0, function* () {
        if (USE_CURRENT_USER_ROLE) {
            yield conf.applyUpdateAsCurrentUser(update);
        }
        else {
            yield conf.applyUpdateViaLambda(update);
        }
    });
    const promise = yield new listr_1.default([
        {
            title: 'load update (grab a coffee)',
            task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                const resp = yield getUpdateWithRetry(tag);
                if (!resp)
                    return;
                const { update, upToDate } = resp;
                ctx.upToDate = upToDate;
                ctx.update = update;
                ctx.willUpdate = force || !upToDate;
            })
        },
        {
            title: 'validate update',
            skip: ctx => !ctx.willUpdate,
            task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                try {
                    yield triggerUpdate(ctx.update);
                }
                catch (err) {
                    if (err.code === 'ValidationError' && err.message.includes('UPDATE_IN_PROGRESS')) {
                        throw new Error('stack is currently updating, please wait for it to finish before applying another update');
                    }
                    throw err;
                }
            })
        },
        {
            title: 'apply update (be patient, or else)',
            skip: ctx => !ctx.willUpdate,
            task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                try {
                    yield conf.waitForStackUpdate(stackId);
                }
                catch (err) {
                    if (err.code === 'ResourceNotReady') {
                        throw new Error('failed to apply update');
                    }
                    throw err;
                }
            })
        }
    ]).run();
    let ctx;
    try {
        ctx = yield promise;
    }
    catch (err) {
        if (errors_1.default.matches(err, errors_2.Errors.NotFound)) {
            throw new Error('failed to fetch the requested update');
        }
        throw err;
    }
    if (ctx.upToDate && !ctx.willUpdate) {
        logger_1.logger.info(`your MyCloud is already up to date!`);
    }
});
const isReleaseCandidateTag = (tag) => /-rc\.\d+$/.test(tag);
const isTransitionReleaseTag = (tag) => /-trans/.test(tag);
const getChoiceTextForTag = (tag) => {
    if (isTransitionReleaseTag(tag)) {
        return `${tag} (transition version, must be applied before the version that follows)`;
    }
    if (isReleaseCandidateTag(tag)) {
        return `${tag} (release candidate with experimental fixes / features)`;
    }
    return tag;
};
//# sourceMappingURL=update.js.map